// ----------------------------------------------------------------------------
// Copyright (C) 2018
//              David Freese, W1HKJ
//
// This file is part of pskmeter.
//
// pskmeter is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// pskmeter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------

#include <cstdlib>
#include <string>
#include <vector>
#include <queue>

#include "config.h"
#include "xml_io.h"
#include "pskmeter_status.h"
#include "pskmeter_main.h"
#include "pskmeter_ui.h"
#include "threads.h"
#include "XmlRpc.h"

using XmlRpc::XmlRpcValue;

pthread_mutex_t mutex_xmlrpc = PTHREAD_MUTEX_INITIALIZER;

std::queue<bool> quePTT;
std::queue<FREQMODE> xml_queue;

static const double TIMEOUT = 1.0;

// these are get only
static const char* main_get_trx_state   = "main.get_trx_state";

// these are set only
static const char* main_set_tx          = "main.tx";
static const char* main_set_rx          = "main.rx";
static const char* rig_set_name         = "rig.set_name";
static const char* rig_set_modes        = "rig.set_modes";
static const char* rig_set_bandwidths   = "rig.set_bandwidths";
static const char* rig_take_control     = "rig.take_control";
static const char* rig_release_control  = "rig.release_control";

// these are get/set
static const char* main_get_frequency   = "main.get_frequency";
static const char* main_set_wf_sideband = "main.set_wf_sideband";
static const char* rig_set_frequency    = "rig.set_frequency";
static const char *rig_set_smeter       = "rig.set_smeter";
static const char *rig_set_pwrmeter     = "rig.set_pwrmeter";
static const char* rig_set_mode         = "rig.set_mode";
static const char* rig_get_mode         = "rig.get_mode";
static const char* rig_set_bandwidth    = "rig.set_bandwidth";
static const char* rig_get_bandwidth    = "rig.get_bandwidth";
static const char* rig_set_notch        = "rig.set_notch";
static const char* rig_get_notch        = "rig.get_notch";

static XmlRpc::XmlRpcClient* client;
static XmlRpcValue* status_query;

#define RIG_UPDATE_INTERVAL  50 // milliseconds
#define CHECK_UPDATE_COUNT   (2000 / RIG_UPDATE_INTERVAL)

bool run_digi_loop = true;
bool fldigi_online = false;

bool ptt_on = false;

int try_count = CHECK_UPDATE_COUNT;

//=====================================================================
// socket ops
//=====================================================================

void open_rig_xmlrpc()
{
	int server_port = atoi(progstatus.server_port.c_str());
	client = new XmlRpc::XmlRpcClient(
				progstatus.server_addr.c_str(),
				server_port );

//	XmlRpc::setVerbosity(5);

	// Create a value suitable for the system.multicall method, which
	// accepts a single argument: an array of struct { methodName, params }
	// where methodName is a string and params is an array of method
	// arguments.  The params member is mandatory so we use a single-element
	// array containing nil.

	XmlRpcValue nil(XmlRpc::nil);
	status_query = new XmlRpcValue;
	const char* status_methods[] = {
		main_get_trx_state, main_get_frequency, rig_get_mode
	};
	for (size_t i = 0; i < sizeof(status_methods)/sizeof(*status_methods); i++) {
		(*status_query)[0][i]["methodName"] = status_methods[i];
		(*status_query)[0][i]["params"][0] = nil;
	};
}

void close_rig_xmlrpc()
{
	guard_lock gl_xmlclose(&mutex_xmlrpc);

	delete client;
	client = NULL;
	delete status_query;
	status_query = NULL;

}

static inline void execute(const char* name, const XmlRpcValue& param, XmlRpcValue& result)
{
	if (client)
		if (!client->execute(name, param, result, TIMEOUT)) {
			fldigi_online = false;
			throw XmlRpc::XmlRpcException(name);
		} else
			fldigi_online = true;
}

// --------------------------------------------------------------------
// send functions
// --------------------------------------------------------------------

void xml_put_modes()
{ 
	if (!fldigi_online) return;

	XmlRpcValue modes, res;
	int i = 0;
	const char *szmodes[] = {"USB", "LSB", "CW", "CWR", "SWU", "SWL", NULL };
	for (const char** mode = szmodes; *mode; mode++, i++) {
		modes[0][i] = *mode;
	}
	try {
		execute(rig_set_modes, modes, res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

void xml_put_name()
{
	if (!fldigi_online) return;
	try {
		XmlRpcValue res;
		execute(rig_set_name, XmlRpcValue(PACKAGE_TARNAME), res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

void xml_put_ptt(bool PTT)
{
	if (!fldigi_online) return;
	try {
		XmlRpcValue res;
		execute((PTT ? main_set_tx : main_set_rx), XmlRpcValue(), res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

void xml_put_freq(int freq)
{
	if (!fldigi_online) return;
	try {
		XmlRpcValue f((double)freq), res;
		execute(rig_set_frequency, f, res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

/*
void send_smeter_val(int val)
{
	if (!fldigi_online) return;
	try {
		XmlRpcValue mval((int)val), res;
		execute(rig_set_smeter, mval, res);
	} catch (const XmlRpc::XmlRpcException& e) {
	}
}

void send_pwrmeter_val(int val)
{
	if (!fldigi_online) return;
	try {
		XmlRpcValue mval((int)val), res;
		execute(rig_set_pwrmeter, mval, res);
	} catch (const XmlRpc::XmlRpcException& e) {
	}
}
*/

void send_queue()
{
	if (!fldigi_online) return;
	while (!xml_queue.empty()) {
		xml_put_freq(xml_queue.front().freq);
		xml_put_mode(xml_queue.front().mode);
		xml_queue.pop();
	}
}

void xml_put_mode(std::string md)
{
	if (!fldigi_online) return;
	try {
		XmlRpcValue mode(md), res;
		execute(rig_set_mode, mode, res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

/*
void send_new_bandwidth(int bw)
{
	if (!selrig->bandwidths_ || !fldigi_online) return;
	try {
		xmlvfo.iBW = bw;
		int selbw = (bw > 0x80) ? (bw >> 8 & 0x7F) : bw;
		XmlRpcValue bandwidth(selrig->bandwidths_[selbw]), res;
		execute(rig_set_bandwidth, bandwidth, res);
	} catch (const XmlRpc::XmlRpcException& e) {
	}
}
*/

void xml_put_sideband()
{
	if (!fldigi_online) return;
	try {
		std::string umodes = "USB CWR SWU";
		XmlRpcValue 
			sideband(
				(umodes.find(Mode_Choice->value()) != std::string::npos) ? 
				"USB" : "LSB"), 
			res;
		execute(main_set_wf_sideband, sideband, res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

// --------------------------------------------------------------------
// receive functions
// --------------------------------------------------------------------

/*
static void check_for_ptt_change(const XmlRpcValue& trx_state)
{
	bool nuptt = (trx_state == "TX");
	if (nuptt != ptt_on) {
		if (XML_DEBUG) {
			string txstate = trx_state;
		}
		ptt_on = nuptt;
		Fl::awake(setPTT, (void*)ptt_on);
	}
}

static void check_for_frequency_change(const XmlRpcValue& freq)
{
	double f = freq;
	long newfreq = (long)f;
	if (newfreq != xmlvfo.freq) {
		xmlvfo.freq = newfreq;
		xmlvfo_changed = true;
	}
}

static bool xmlmode_changed = false;

static void check_for_mode_change(const XmlRpcValue& new_mode)
{
	if (!selrig->modes_) 
		return;
	string smode = new_mode;
	xmlmode_changed = false;
	if (smode != selrig->modes_[xmlvfo.imode]) {
		int imode = 0;
		while (selrig->modes_[imode] != NULL && smode != selrig->modes_[imode])
			imode++;
		if (selrig->modes_[imode] != NULL && imode != xmlvfo.imode) {
			xmlvfo.imode = imode;
			xmlvfo_changed = true;
			xmlmode_changed = true;
		}
	}
}

*/

void set_fldigi_connect (void *d)
{
	bool b = (long *)d;
	box_online->color(b ? FL_GREEN : FL_BACKGROUND2_COLOR);
	box_online->redraw();
}

void xml_put_info()
{
	XmlRpcValue res;
	try {
		execute(rig_take_control, XmlRpcValue(), res);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw e;
	}
	try {
		xml_put_name();
		xml_put_modes();
		Fl::awake(set_fldigi_connect, (void *)1);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw;
	}
}

static void get_fldigi_status()
{
/*
	XmlRpcValue status;
	XmlRpcValue param((int)0);
	string xmlcall;
	try {
		xmlcall = "system.multicall";
		execute(xmlcall.c_str(), *status_query, status);
		check_for_ptt_change(status[0][0]);
		if (!ptt_on) {
			xmlvfo.src = XML;
			xmlvfo_changed = false;
			check_for_frequency_change(status[1][0]);
			check_for_mode_change(status[2][0]);
			check_for_bandwidth_change(status[3][0]);
			if (xmlvfo_changed)
				push2que();
		}
	} catch (const XmlRpc::XmlRpcException& e) {
		throw e;
	}
*/
}

void digi_loop(void *)
//void * digi_loop(void *d)
{
//	for (;;) {
//		MilliSleep(RIG_UPDATE_INTERVAL);

		if (!fldigi_online) {
			if (--try_count == 0) {
				try {
					guard_lock gl_xml(&mutex_xmlrpc);
					xml_put_info();
				} catch (const XmlRpc::XmlRpcException& e) {
//std::cout << "digi_loop: " << e.getMessage() << std::endl;
					try_count = CHECK_UPDATE_COUNT;
					Fl::awake(set_fldigi_connect, (void *)0);
				}
			}
		} else {
			try {
				guard_lock gl_xml(&mutex_xmlrpc);
				get_fldigi_status();
				send_queue();
			} catch (const XmlRpc::XmlRpcException& e) {
//std::cout << "digi_loop: " << e.getMessage() << std::endl;
				try_count = CHECK_UPDATE_COUNT;
				Fl::awake(set_fldigi_connect, (void *)0);
			}
		}
//	}
//	return NULL;
	Fl::repeat_timeout(0.1, digi_loop);
}

